unit SSLOpenSSL;

interface

uses
  Classes,
  IdException,
  IdStackConsts,
  SSLIntercept,
  IdSocketHandle,
  SSLOpenSSLHeaders,
  IdGlobal,
  IdTCPServer,
  IdIntercept, SysUtils;

type
  TX509 = class;

  TSSLVersion = (sslvSSLv2, sslvSSLv23, sslvSSLv3, sslvTLSv1);
  TSSLMode = (sslmUnassigned, sslmClient, sslmServer, sslmBoth);
  TSSLVerifyMode = (sslvrfPeer, sslvrfFailIfNoPeerCert, sslvrfClientOnce);
  TSSLVerifyModeSet = set of TSSLVerifyMode;
  TSSLErrorMode = (sslemClient, sslemServer);
  TSSLAction = (sslRead, sslWrite);

  TULong = packed record
    case Byte of
      0: (B1, B2, B3, B4: Byte);
      1: (W1, W2: Word);
      2: (L1: Longint);
      3: (C1: Cardinal);
  end;

  TEVP_MD = record
    Length: Integer;
    MD: array[0..OPENSSL_EVP_MAX_MD_SIZE - 1] of Char;
  end;

  TByteArray = record
    Length: Integer;
    Data: PChar;
  end;

  TCallbackEvent = procedure(Msg: string) of object;
  TPasswordEvent = procedure(var Password: string) of object;
  TVerifyPeerEvent = function(Certificate: TX509; Ok: Integer): Boolean of object;

  TSSLOptions = class(TPersistent)
  protected
    fsRootCertFile, fsCertFile, fsKeyFile: TFileName;
    fMethod: TSSLVersion;
    fMode: TSSLMode;

    fVerifyDepth: Integer;
    fVerifyMode: TSSLVerifyModeSet;
    fVerifyFile, fVerifyDirs, fCipherList: string;
  published
    property RootCertFile: TFileName read fsRootCertFile write fsRootCertFile;
    property CertFile: TFileName read fsCertFile write fsCertFile;
    property KeyFile: TFileName read fsKeyFile write fsKeyFile;
    property Method: TSSLVersion read fMethod write fMethod;
    property Mode: TSSLMode read fMode write fMode;
    property VerifyMode: TSSLVerifyModeSet read fVerifyMode write fVerifyMode;
    property VerifyDepth: Integer read fVerifyDepth write fVerifyDepth;
  end;

  TSSLContext = class(TObject)
  protected
    fMethod: TSSLVersion;
    fMode: TSSLMode;
    fsRootCertFile, fsCertFile, fsKeyFile: string;
    fVerifyDepth: Integer;
    fVerifyMode: TSSLVerifyModeSet;
    fContext: PSSL_CTX;
    fStatusInfoOn: Boolean;
    fVerifyOn: Boolean;
    procedure DestroyContext;
    function InternalGetMethod: PSSL_METHOD;
    procedure SetMode(const Value: TSSLMode);

    procedure SetVerifyMode(Mode: TSSLVerifyModeSet);
    function GetVerifyMode: TSSLVerifyModeSet;

    function LoadOpenSLLibrary: Boolean;
    procedure UnLoadOpenSLLibrary;
  public
    Parent: TObject;
    constructor Create;
    destructor Destroy; override;
    function LoadRootCert: Boolean;
    function LoadCert: Boolean;
    function LoadKey: Boolean;
    property StatusInfoOn: Boolean read fStatusInfoOn write fStatusInfoOn;
    property VerifyOn: Boolean read fVerifyOn write fVerifyOn;
  published
    property Method: TSSLVersion read fMethod write fMethod;
    property Mode: TSSLMode read fMode write SetMode;
    property RootCertFile: string read fsRootCertFile write fsRootCertFile;
    property CertFile: string read fsCertFile write fsCertFile;
    property KeyFile: string read fsKeyFile write fsKeyFile;
    property VerifyMode: TSSLVerifyModeSet read GetVerifyMode write
      SetVerifyMode;
    property VerifyDepth: Integer read fVerifyDepth write fVerifyDepth;
  end;

  TSSLSocket = class(TObject)
  private
    fPeerCert: TX509;

    fParent: TObject;
    function GetPeerCert: TX509;
    function GetSSLError(retCode: Integer; DataLen: Integer; Mode:
      TSSLErrorMode; Action: TSSLAction): Integer;
  public
    fSSL: PSSL;

    constructor Create(Parent: TObject);
    procedure Accept(const pHandle: TIdStackSocketHandle; fSSLContext:
      TSSLContext);
    procedure Connect(const pHandle: TIdStackSocketHandle; fSSLContext:
      TSSLContext);
    destructor Destroy; override;
    function GetSessionID: TByteArray;
    function GetSessionIDAsString: string;
    procedure SetCipherList(CipherList: string);

    property PeerCert: TX509 read GetPeerCert;
  end;

  TConnectionInterceptOpenSSL = class(TSSLConnectionIntercept)
  private
    fSSLContext: TSSLContext;
    fxSSLOptions: TSSLOptions;
    fSSLSocket: TSSLSocket;
    fIsPeer: Boolean;

    fOnStatusInfo: TCallbackEvent;
    fOnGetPassword: TPasswordEvent;
    fOnVerifyPeer: TVerifyPeerEvent;
    fSSLLayerClosed: Boolean;
    function GetPeerCert: TX509;
    procedure CreateSSLContext(axMode: TSSLMode);

  protected
    procedure DoStatusInfo(Msg: string); virtual;
    procedure DoGetPassword(var Password: string); virtual;
    function DoVerifyPeer(Certificate: TX509; Ok: Integer): Boolean; virtual;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Connect(ABinding: TIdSocketHandle); override;
    procedure Disconnect; override;
    function Recv(var ABuf; ALen: Integer): Integer; override;
    function Send(var ABuf; ALen: Integer): Integer; override;
    property PeerCert: TX509 read GetPeerCert;
  published
    property SSLOptions: TSSLOptions read fxSSLOptions write fxSSLOptions;
    property OnStatusInfo: TCallbackEvent read fOnStatusInfo write
      fOnStatusInfo;
    property OnGetPassword: TPasswordEvent read fOnGetPassword write
      fOnGetPassword;
    property OnVerifyPeer: TVerifyPeerEvent read fOnVerifyPeer write
      fOnVerifyPeer;
  end;

  TServerInterceptOpenSSL = class(TSSLServerIntercept)
  private
    fSSLContext: TSSLContext;
    fxSSLOptions: TSSLOptions;
    fIsInitialized: Boolean;
    fOnStatusInfo: TCallbackEvent;
    fOnGetPassword: TPasswordEvent;
    fOnVerifyPeer: TVerifyPeerEvent;
    procedure CreateSSLContext(axMode: TSSLMode);
  protected
    procedure DoStatusInfo(Msg: string); virtual;
    procedure DoGetPassword(var Password: string); virtual;
    function DoVerifyPeer(Certificate: TX509; Ok: Integer): Boolean; virtual;
  public
    procedure Init; override;
    function Accept(ABinding: TIdSocketHandle): TIdConnectionIntercept;
      override;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  published
    property SSLOptions: TSSLOptions read fxSSLOptions write fxSSLOptions;
    property OnStatusInfo: TCallbackEvent read fOnStatusInfo write
      fOnStatusInfo;
    property OnGetPassword: TPasswordEvent read fOnGetPassword write
      fOnGetPassword;
    property OnVerifyPeer: TVerifyPeerEvent read fOnVerifyPeer write
      fOnVerifyPeer;
  end;

  TX509Name = class(TObject)
  private
    fX509Name: PX509_NAME;
    function CertInOneLine: string;
    function GetHash: TULong;
    function GetHashAsString: string;
  public
    constructor Create(aX509Name: PX509_NAME);

    property Hash: TULong read GetHash;
    property HashAsString: string read GetHashAsString;
    property OneLine: string read CertInOneLine;
  end;

  TX509 = class(TObject)
  protected
    FX509: PX509;
    FSubject: TX509Name;
    FIssuer: TX509Name;
    function RSubject: TX509Name;
    function RIssuer: TX509Name;
    function RnotBefore: TDateTime;
    function RnotAfter: TDateTime;
    function RFingerprint: TEVP_MD;
    function RFingerprintAsString: string;
  public
    constructor Create(aX509: PX509); virtual;
    destructor Destroy; override;

    property Fingerprint: TEVP_MD read RFingerprint;
    property FingerprintAsString: string read RFingerprintAsString;
    property Subject: TX509Name read RSubject;
    property Issuer: TX509Name read RIssuer;
    property notBefore: TDateTime read RnotBefore;
    property notAfter: TDateTime read RnotAfter;
  end;

type
  EIdOpenSSLError = class(EIdException);
  EIdOpenSSLLoadError = class(EIdOpenSSLError);
  EIdOSSLCouldNotLoadSSLLibrary = class(EIdOpenSSLLoadError);
  EIdOSSLModeNotSet = class(EIdOpenSSLError);
  EIdOSSLGetMethodError = class(EIdOpenSSLError);
  EIdOSSLCreatingContextError = class(EIdOpenSSLError);
  EIdOSSLLoadingRootCertError = class(EIdOpenSSLLoadError);
  EIdOSSLLoadingCertError = class(EIdOpenSSLLoadError);
  EIdOSSLLoadingKeyError = class(EIdOpenSSLLoadError);
  EIdOSSLSettingCipherError = class(EIdOpenSSLError);
  EIdOSSLDataBindingError = class(EIdOpenSSLError);
  EIdOSSLAcceptError = class(EIdOpenSSLError);
  EIdOSSLConnectError = class(EIdOpenSSLError);

function LogicalAnd(A, B: Integer): Boolean;
procedure InfoCallback(sslSocket: PSSL; where: Integer; ret: Integer); cdecl;
function PasswordCallback(buf: PChar; size: Integer; rwflag: Integer; userdata:
  Pointer): Integer; cdecl;
function VerifyCallback(Ok: Integer; ctx: PX509_STORE_CTX): Integer; cdecl;

implementation

uses
  IdResourceStrings,
  Windows;

var
  DLLLoadCount: Integer = 0;

function LoadOpenSLLibrary: Boolean;
begin

  SslRandScreen;
  SslLoadErrorStrings;

  result := SslAddSslAlgorithms > 0;
end;

procedure UnLoadOpenSLLibrary;
begin
////////////////////////////////
end;

constructor TServerInterceptOpenSSL.Create(AOwner: TComponent);
begin
  inherited;
  fIsInitialized := False;
  fxSSLOptions := TSSLOptions.Create;
end;

destructor TServerInterceptOpenSSL.Destroy;
begin
  if fSSLContext <> nil then
  begin
    FreeAndNil(fSSLContext);
  end;
  fxSSLOptions.Destroy;
  inherited;
end;

procedure TServerInterceptOpenSSL.Init;
begin
  CreateSSLContext(sslmServer);
  fIsInitialized := True;
end;

function TServerInterceptOpenSSL.Accept(ABinding: TIdSocketHandle):
  TIdConnectionIntercept;
var
  tmpIdCIOpenSSL: TConnectionInterceptOpenSSL;
begin
  if not fIsInitialized then
  begin
    Init;
  end;
  tmpIdCIOpenSSL := TConnectionInterceptOpenSSL.Create(self);
  tmpIdCIOpenSSL.fIsPeer := True;
  tmpIdCIOpenSSL.fxSSLOptions := fxSSLOptions;
  tmpIdCIOpenSSL.fSSLSocket := TSSLSocket.Create(self);
  tmpIdCIOpenSSL.fSSLSocket.Accept(ABinding.Handle, fSSLContext);

  Result := tmpIdCIOpenSSL;
end;

procedure TServerInterceptOpenSSL.CreateSSLContext(axMode: TSSLMode);
begin
  fSSLContext := TSSLContext.Create;
  with fSSLContext do
  begin
    Parent := self;
    RootCertFile := SSLOptions.RootCertFile;
    CertFile := SSLOptions.CertFile;
    KeyFile := SSLOptions.KeyFile;
    if Assigned(fOnStatusInfo) then
    begin
      StatusInfoOn := True;
    end
    else
    begin
      StatusInfoOn := False;
    end;

    VerifyMode := SSLOptions.VerifyMode;
    if Assigned(fOnVerifyPeer) then
    begin
      VerifyOn := True;
    end
    else
    begin
      VerifyOn := False;
    end;

    Mode := axMode;
    Method := SSLOptions.Method;
  end;
end;

procedure TServerInterceptOpenSSL.DoStatusInfo(Msg: string);
begin
  if Assigned(fOnStatusInfo) then
    fOnStatusInfo(Msg);
end;

procedure TServerInterceptOpenSSL.DoGetPassword(var Password: string);
begin
  if Assigned(fOnGetPassword) then
    fOnGetPassword(Password);
end;

function TServerInterceptOpenSSL.DoVerifyPeer(Certificate: TX509; Ok: Integer)
		: Boolean;
begin
  Result := True;
  if Assigned(fOnVerifyPeer) then
    Result := fOnVerifyPeer(Certificate, Ok);
end;

constructor TConnectionInterceptOpenSSL.Create(AOwner: TComponent);
begin
  inherited;
  fIsPeer := False;
  fxSSLOptions := TSSLOptions.Create;
  fSSLLayerClosed := True;
end;

destructor TConnectionInterceptOpenSSL.Destroy;
begin
  FreeAndNil(fSSLSocket);
  if not fIsPeer then
  begin
    FreeAndNil(fSSLContext);
    fxSSLOptions.Destroy;
  end;
  inherited;
end;

procedure TConnectionInterceptOpenSSL.Connect(ABinding: TIdSocketHandle);
begin
  inherited;
  CreateSSLContext(sslmClient);

  fSSLSocket := TSSLSocket.Create(self);
  if fSSLSocket <> nil then
  begin
    fSSLSocket.Connect(ABinding.Handle, fSSLContext);
  end
  else
  begin
  end;
end;

procedure TConnectionInterceptOpenSSL.Disconnect;
begin
  FreeAndNil(fSSLSocket);
  if not fIsPeer then
  begin
    FreeAndNil(fSSLContext);
  end;
  inherited;
end;

function TConnectionInterceptOpenSSL.Recv(var ABuf; ALen: Integer): Integer;
begin
  Result := SslRead_1(fSSLSocket.fSSL, @ABuf, ALen);
  while (fSSLSocket.GetSSLError(Result, ALen, sslemClient, sslRead) =
    OPENSSL_SSL_ERROR_WANT_READ) do
  begin
    Result := SslRead_1(fSSLSocket.fSSL, @ABuf, ALen);
  end;
end;

function TConnectionInterceptOpenSSL.Send(var ABuf; ALen: Integer): Integer;
begin
  Result := SslWrite_1(fSSLSocket.fSSL, @ABuf, ALen);
  while (fSSLSocket.GetSSLError(Result, ALen, sslemClient, sslWrite) =
    OPENSSL_SSL_ERROR_WANT_WRITE) do
  begin
    Result := SslWrite_1(fSSLSocket.fSSL, @ABuf, ALen);
  end;
end;

function TSSLSocket.GetSSLError(retCode: Integer; DataLen: Integer; Mode:
  TSSLErrorMode; Action: TSSLAction): Integer;
begin
  Result := SslGetError(fSSL, retCode);
  case Result of
    OPENSSL_SSL_ERROR_NONE:
      Result := OPENSSL_SSL_ERROR_NONE;
    OPENSSL_SSL_ERROR_WANT_WRITE:
      Result := OPENSSL_SSL_ERROR_WANT_WRITE;
    OPENSSL_SSL_ERROR_WANT_READ:
      Result := OPENSSL_SSL_ERROR_WANT_READ;
    OPENSSL_SSL_ERROR_ZERO_RETURN:
      Result := OPENSSL_SSL_ERROR_NONE;
    OPENSSL_SSL_ERROR_SYSCALL:
      Result := OPENSSL_SSL_ERROR_NONE;
    OPENSSL_SSL_ERROR_SSL:
      Result := OPENSSL_SSL_ERROR_NONE;
  end;
end;

procedure TConnectionInterceptOpenSSL.CreateSSLContext(axMode: TSSLMode);
begin
  fSSLContext := TSSLContext.Create;
  with fSSLContext do
  begin
    Parent := self;
    RootCertFile := SSLOptions.RootCertFile;
    CertFile := SSLOptions.CertFile;
    KeyFile := SSLOptions.KeyFile;
    if Assigned(fOnStatusInfo) then
    begin
      StatusInfoOn := True;
    end
    else
    begin
      StatusInfoOn := False;
    end;
    if Assigned(fOnVerifyPeer) then
    begin
      VerifyOn := True;
    end
    else
    begin
      VerifyOn := False;
    end;
    Method := SSLOptions.Method;
    Mode := axMode;
  end;
end;

function TConnectionInterceptOpenSSL.GetPeerCert: TX509;
begin
  if fSSLContext <> nil then
  begin
    Result := fSSLSocket.PeerCert;
  end
  else
  begin
    Result := nil;
  end;
end;

procedure TConnectionInterceptOpenSSL.DoStatusInfo(Msg: string);
begin
  if Assigned(fOnStatusInfo) then
    fOnStatusInfo(Msg);
end;

procedure TConnectionInterceptOpenSSL.DoGetPassword(var Password: string);
begin
  if Assigned(fOnGetPassword) then
    fOnGetPassword(Password);
end;

function TConnectionInterceptOpenSSL.DoVerifyPeer(Certificate: TX509; Ok: Integer):
  Boolean;
begin
  Result := True;
  if Assigned(fOnVerifyPeer) then
    Result := fOnVerifyPeer(Certificate, Ok);
end;

function PasswordCallback(buf: PChar; size: Integer; rwflag: Integer; userdata:
  Pointer): Integer; cdecl;
var
  Password: string;
  IdSSLContext: TSSLContext;
begin
  Password := '';

  IdSSLContext := TSSLContext(userdata);

  if (IdSSLContext.Parent is TServerInterceptOpenSSL) then
  begin
    (IdSSLContext.Parent as TServerInterceptOpenSSL).DoGetPassword(Password);
  end;
  if (IdSSLContext.Parent is TConnectionInterceptOpenSSL) then
  begin
    (IdSSLContext.Parent as
      TConnectionInterceptOpenSSL).DoGetPassword(Password);
  end;

  Password := Copy(Password, 1, size);
  StrLCopy(buf, PChar(Password), size);
  Result := Length(Password);
end;

procedure InfoCallback(sslSocket: PSSL; where: Integer; ret: Integer); cdecl;
var
  IdSSLSocket: TSSLSocket;
  StatusStr: string;
begin
  IdSSLSocket := TSSLSocket(SslGetAppData(sslSocket));

  StatusStr := Format(RSOSSLStatusString,
    [StrPas(SslStateStringLong(sslSocket))]);
  if (IdSSLSocket.fParent is TServerInterceptOpenSSL) then
  begin
    (IdSSLSocket.fParent as TServerInterceptOpenSSL).DoStatusInfo(StatusStr);
  end;
  if (IdSSLSocket.fParent is TConnectionInterceptOpenSSL) then
  begin
    (IdSSLSocket.fParent as
      TConnectionInterceptOpenSSL).DoStatusInfo(StatusStr);
  end;

end;

function AddMins(const DT: TDateTime; const Mins: Extended): TDateTime;
begin
  Result := DT + Mins / (60 * 24)
end;

function AddHrs(const DT: TDateTime; const Hrs: Extended): TDateTime;
begin
  Result := DT + Hrs / 24.0
end;

function GetLocalTZBias: LongInt;
var
  TZ: TTimeZoneInformation;
begin
  case GetTimeZoneInformation(TZ) of
    TIME_ZONE_ID_STANDARD: Result := TZ.Bias + TZ.StandardBias;
    TIME_ZONE_ID_DAYLIGHT: Result := TZ.Bias + TZ.DaylightBias;
  else
    Result := TZ.Bias;
  end;
end;

function GetLocalTime(const DT: TDateTime): TDateTime;
begin
  Result := DT - GetLocalTZBias / (24 * 60);
end;

constructor TSSLContext.Create;
begin
  inherited;

  if DLLLoadCount <= 0 then
  begin
    if not SSLOpenSSL.LoadOpenSLLibrary then
    begin
      raise EIdOSSLCouldNotLoadSSLLibrary.Create(RSOSSLCouldNotLoadSSLLibrary);
    end;
  end;
  Inc(DLLLoadCount);

  fMode := sslmUnassigned;
end;

destructor TSSLContext.Destroy;
begin
  DestroyContext;
  inherited;
end;

procedure TSSLContext.DestroyContext;
begin
  if fContext <> nil then
  begin
    SslCtxFree(fContext);
    fContext := nil;
  end;
end;

function LogicalAnd(A, B: Integer): Boolean;
begin
  Result := (A and B) = B;
end;

function VerifyCallback(Ok: Integer; ctx: PX509_STORE_CTX): Integer; cdecl;
var
  hcert: PX509;
  Certificate: TX509;
  hSSL: PSSL;
  IdSSLSocket: TSSLSocket;
  VerifiedOK: Boolean;
begin
  VerifiedOK := False;
  hcert := SslX509StoreCtxGetCurrentCert(ctx);
  hSSL := SslX509StoreCtxGetAppData(ctx);
  Certificate := TX509.Create(hcert);
  if hSSL <> nil then
    IdSSLSocket := TSSLSocket(SslGetAppData(hSSL))
  else
    IdSSLSocket := nil;
  SslX509StoreCtxGetError(ctx);
  SslX509StoreCtxGetErrorDepth(ctx);

  if (IdSSLSocket.fParent is TServerInterceptOpenSSL) then
  begin
    VerifiedOK := (IdSSLSocket.fParent as
      TServerInterceptOpenSSL).DoVerifyPeer(Certificate, Ok);
  end;
  if (IdSSLSocket.fParent is TConnectionInterceptOpenSSL) then
  begin
    VerifiedOK := (IdSSLSocket.fParent as
      TConnectionInterceptOpenSSL).DoVerifyPeer(Certificate, Ok);
  end;

  Certificate.Destroy;
  if VerifiedOK and (Ok > 0) then
  begin
    Result := 1;//0;
  end
  else
  begin
    Result := 0;//-1;
  end;
end;

function TranslateInternalVerifyToSLL(Mode: TSSLVerifyModeSet): Integer;
begin
  Result := OPENSSL_SSL_VERIFY_NONE;
  if sslvrfPeer in Mode then Result := Result or OPENSSL_SSL_VERIFY_PEER;
  if sslvrfFailIfNoPeerCert in Mode then
    Result := Result or OPENSSL_SSL_VERIFY_FAIL_IF_NO_PEER_CERT;
  if sslvrfClientOnce in Mode then
    Result := Result or OPENSSL_SSL_VERIFY_CLIENT_ONCE;
end;

procedure TSSLContext.SetVerifyMode(Mode: TSSLVerifyModeSet);
begin
  fVerifyMode := Mode;
  if fContext <> nil then
  begin
    SslCtxSetVerify(fContext, TranslateInternalVerifyToSLL(fVerifyMode),
      PFunction(@VerifyCallback));
  end;
end;

function TSSLContext.GetVerifyMode: TSSLVerifyModeSet;
begin
  Result := fVerifyMode;
end;

function TSSLContext.InternalGetMethod: PSSL_METHOD;
begin
  if fMode = sslmUnassigned then
  begin
    raise EIdOSSLModeNotSet.create(RSOSSLModeNotSet);
  end;
  case fMethod of
    sslvSSLv2:
      case fMode of
        sslmServer: Result := SslMethodServerV2;
        sslmClient: Result := SslMethodClientV2;
        sslmBoth: Result := SslMethodV2;
      else
        Result := SslMethodV2;
      end;

    sslvSSLv23:
      case fMode of
        sslmServer: Result := SslMethodServerV23;
        sslmClient: Result := SslMethodClientV23;
        sslmBoth: Result := SslMethodV23;
      else
        Result := SslMethodV23;
      end;

    sslvSSLv3:
      case fMode of
        sslmServer: Result := SslMethodServerV3;
        sslmClient: Result := SslMethodClientV3;
        sslmBoth: Result := SslMethodV3;
      else
        Result := SslMethodV3;
      end;

    sslvTLSv1:
      case fMode of
        sslmServer: Result := SslMethodServerTLSV1;
        sslmClient: Result := SslMethodClientTLSV1;
        sslmBoth: Result := SslMethodTLSV1;
      else
        Result := SslMethodTLSV1;
      end;
  else
    raise EIdOSSLGetMethodError.Create(RSSSLGetMethodError);
  end;
end;

function TSSLContext.LoadRootCert: Boolean;
var
  pStr: PChar;
  error: Integer;
begin
  pStr := StrNew(PChar(RootCertFile));
  error := SslCtxLoadVerifyLocations(
    fContext,
    pStr,
    nil);
  if error <= 0 then
  begin
    Result := False
  end
  else
  begin
    Result := True;
  end;

  StrDispose(pStr);
end;

function TSSLContext.LoadCert: Boolean;
var
  pStr: PChar;
  error: Integer;
begin
  pStr := StrNew(PChar(CertFile));
  error := SslCtxUseCertificateFile(
    fContext,
    pStr,
    OPENSSL_SSL_FILETYPE_PEM);
  if error <= 0 then
    Result := False
  else
    Result := True;

  StrDispose(pStr);
end;

function TSSLContext.LoadKey: Boolean;
var
  pStr: PChar;
  error: Integer;
begin
  Result := True;

  pStr := StrNew(PChar(fsKeyFile));
  error := SslCtxUsePrivateKeyFile(
    fContext,
    pStr,
    OPENSSL_SSL_FILETYPE_PEM);

  if error <= 0 then
  begin
    Result := False;
  end
  else
  begin
    error := SslCtxCheckPrivateKeyFile(fContext);
    if error <= 0 then
    begin
      Result := False;
    end;
  end;

  StrDispose(pStr);
end;

function TSSLContext.LoadOpenSLLibrary: Boolean;
begin

  SslRandScreen;
  SslLoadErrorStrings;

  result := SslAddSslAlgorithms > 0;
end;

procedure TSSLContext.SetMode(const Value: TSSLMode);
var
  method: PSSL_METHOD;
  error: Integer;
begin
  if fMode = Value then
  begin
    exit;
  end;
  fMode := Value;

  DestroyContext;

  if fMode <> sslmUnassigned then
  begin
    method := InternalGetMethod;

    fContext := SslCtxNew(method);
    if fContext = nil then
    begin
      raise EIdOSSLCreatingContextError.Create(RSSSLCreatingContextError);
    end;

    SslCtxSetDefaultPasswdCb(fContext, @PasswordCallback);
    SslCtxSetDefaultPasswdCbUserdata(fContext, self);

    if RootCertFile <> '' then
    begin
      if not LoadRootCert then
      begin
        raise EIdOSSLLoadingRootCertError.Create(RSSSLLoadingRootCertError);
      end;
    end;

    if CertFile <> '' then
    begin
      if not LoadCert then
      begin
        raise EIdOSSLLoadingCertError.Create(RSSSLLoadingCertError);
      end;
    end;

    if KeyFile <> '' then
    begin
      if not LoadKey then
      begin
        raise EIdOSSLLoadingKeyError.Create(RSSSLLoadingKeyError);
      end;
    end;

    if StatusInfoOn then
    begin
      SslCtxSetInfoCallback(fContext, PFunction(@InfoCallback));
    end;

    error := SslCtxSetCipherList(fContext, OPENSSL_SSL_DEFAULT_CIPHER_LIST);
    if error <= 0 then
    begin
      raise EIdOSSLSettingCipherError.Create(RSSSLSettingChiperError);
    end;

    if VerifyOn then
    begin
      SetVerifyMode(fVerifyMode);
    end;
  end;
end;

procedure TSSLContext.UnLoadOpenSLLibrary;
begin
////////////////////////////////////
//  SSLOpenSSLHeaders.Unload;
end;

constructor TSSLSocket.Create(Parent: TObject);
begin
  fParent := Parent;
end;

destructor TSSLSocket.Destroy;
begin
  if fSSL <> nil then
  begin
    SslSetShutdown(fSSL, OPENSSL_SSL_SENT_SHUTDOWN);
    SslShutdown(fSSL);
    SslFree(fSSL);
    fSSL := nil;
  end;
  inherited;
end;

procedure TSSLSocket.Accept(const pHandle: TIdStackSocketHandle; fSSLContext:
  TSSLContext);
var
  err: Integer;
begin
  fSSL := SslNew(fSSLContext.fContext);
  if fSSL = nil then exit;

  if SslSetAppData(fSSL, self) <= 0 then
  begin
    raise EIdOSSLDataBindingError.Create(RSSSLDataBindingError);
    exit;
  end;

  SslSetFd(fSSL, pHandle);
  err := SslAccept(fSSL);
  if err <= 0 then
  begin
    raise EIdOSSLAcceptError.Create(RSSSLAcceptError);
  end;
end;

procedure TSSLSocket.Connect(const pHandle: TIdStackSocketHandle; fSSLContext:
  TSSLContext);
var
  error: Integer;
begin
  fSSL := SslNew(fSSLContext.fContext);
  if fSSL = nil then exit;

  if SslSetAppData(fSSL, self) <= 0 then
  begin
    raise EIdOSSLDataBindingError.Create(RSSSLDataBindingError);
    exit;
  end;

  SslSetFd(fSSL, pHandle);
  error := SslConnect(fSSL);
  if error <= 0 then
  begin
    raise EIdOSSLConnectError.Create(RSSSLConnectError);
  end;
end;

function TSSLSocket.GetPeerCert: TX509;
var
  X509: PX509;
begin
  if fPeerCert = nil then
  begin
    X509 := SslGetPeerCertificate(fSSL);
    if X509 <> nil then
    begin
      fPeerCert := TX509.Create(X509);
    end;
  end;
  Result := fPeerCert;
end;

function TSSLSocket.GetSessionID: TByteArray;
var
  pSession: PSSL_SESSION;
  tmpArray: TByteArray;
begin
  Result.Length := 0;
  FillChar(tmpArray, SizeOf(TByteArray), 0);
  if fSSL <> nil then
  begin
    pSession := SslGetSession(fSSL);
    if pSession <> nil then
    begin
      SslSessionGetId(pSession, @tmpArray.Data, @tmpArray.Length);
      Result := tmpArray;
    end;
  end;
end;

function TSSLSocket.GetSessionIDAsString: string;
var
  Data: TByteArray;
  i: Integer;
begin
  Result := '';
  Data := GetSessionID;
  for i := 0 to Data.Length - 1 do
  begin
    Result := Result + Format('%.2x', [Byte(Data.Data[I])]); {do not localize}
  end;
end;

procedure TSSLSocket.SetCipherList(CipherList: string);
begin
end;

function TX509Name.CertInOneLine: string;
var
  OneLine: array[0..2048] of Char;
begin
  if FX509Name = nil then
  begin
    Result := '';
  end
  else
  begin
    Result := StrPas(SslX509NameOneline(FX509Name, PChar(@OneLine),
      sizeof(OneLine)));
  end;
end;

function TX509Name.GetHash: TULong;
begin
  if FX509Name = nil then
  begin
    FillChar(Result, SizeOf(Result), 0)
  end
  else
  begin
    Result.C1 := SslX509NameHash(FX509Name);
  end;
end;

function TX509Name.GetHashAsString: string;
begin
  Result := Format('%.8x', [Hash.L1]); {do not localize}
end;

constructor TX509Name.Create(aX509Name: PX509_NAME);
begin
  inherited Create;

  FX509Name := aX509Name;
end;

constructor TX509.Create(aX509: PX509);
begin
  inherited Create;

  FX509 := aX509;
  FSubject := nil;
  FIssuer := nil;
end;

destructor TX509.Destroy;
begin
  if Assigned(FSubject) then FSubject.Destroy;
  if Assigned(FIssuer) then FIssuer.Destroy;

  inherited Destroy;
end;

function TX509.RSubject: TX509Name;
var
  x509_name: PX509_NAME;
begin
  if not Assigned(FSubject) then
  begin
    if FX509 <> nil then
      x509_name := SslX509GetSubjectName(FX509)
    else
      x509_name := nil;
    FSubject := TX509Name.Create(x509_name);
  end;
  Result := FSubject;
end;

function TX509.RIssuer: TX509Name;
var
  x509_name: PX509_NAME;
begin
  if not Assigned(FIssuer) then
  begin
    if FX509 <> nil then
      x509_name := SslX509GetIssuerName(FX509)
    else
      x509_name := nil;
    FIssuer := TX509Name.Create(x509_name);
  end;
  Result := FIssuer;
end;

function TX509.RFingerprint: TEVP_MD;
begin
  SslX509Digest(FX509, SslEvpMd5, PChar(@Result.MD), @Result.Length);
end;

function TX509.RFingerprintAsString: string;
var
  I: Integer;
  EVP_MD: TEVP_MD;
begin
  Result := '';
  EVP_MD := Fingerprint;
  for I := 0 to EVP_MD.Length - 1 do
  begin
    if I <> 0 then Result := Result + ':';
    Result := Result + Format('%.2x', [Byte(EVP_MD.MD[I])]); {do not localize}
  end;
end;

function UTCTime2DateTime(UCTTime: PASN1_UTCTIME): TDateTime;
var
  year: Word;
  month: Word;
  day: Word;
  hour: Word;
  min: Word;
  sec: Word;
  tz_h: Integer;
  tz_m: Integer;
begin
  Result := 0;
  if SslUCTTimeDecode(UCTTime, PUShort(@year), PUShort(@month), PUShort(@day),
    PUShort(@hour)
    , PUShort(@min), PUShort(@sec), SSLOpenSSLHeaders.PInteger(@tz_h)
    , SSLOpenSSLHeaders.PInteger(@tz_m)) > 0 then
  begin
    Result := EncodeDate(year, month, day) + EncodeTime(hour, min, sec, 0);
    AddMins(Result, tz_m);
    AddHrs(Result, tz_h);
    Result := GetLocalTime(Result);
  end;
end;

function TX509.RnotBefore: TDateTime;
begin
  if FX509 = nil then
    Result := 0
  else
    Result := UTCTime2DateTime(SslX509GetNotBefore(FX509));
end;

function TX509.RnotAfter: TDateTime;
begin
  if FX509 = nil then
    Result := 0
  else
    Result := UTCTime2DateTime(SslX509GetNotAfter(FX509));
end;

initialization

finalization
  UnLoadOpenSLLibrary;

end.

