//---------------------------------------------------------------------------
/*
Copyright (c) 2013 Serge L. Ryadkow admin@opcgate.ru http://opcgate.ru 

Permission is hereby granted, free of charge, to any person obtaining a copy of this 
software and associated documentation files (the "Software"), to deal in the Software 
without restriction, including without limitation the rights to use, copy, modify, 
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to 
permit persons to whom the Software is furnished to do so, subject to the following 
conditions:

The above copyright notice and this permission notice shall be included in all copies 
or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR 
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE 
FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR 
OTHER DEALINGS IN THE SOFTWARE. 
*/
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
#include <vcl.h>

#pragma hdrstop

#include <IndyStatSSL.h>
#include <SHA2Str.h>

#include "Gate.h"
#include "Protocol.h"
#include "Session.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
//---------------------------------------------------------------------------
namespace OPCGate {
//---------------------------------------------------------------------------
__fastcall TOPCGate::TOPCGate(int port, TLogFlags logFlags,
		int connTOms, int subsTOms, bool buffSubs)
  :	LogFlags(logFlags)
{
	::LibInit();
  Intercept = 0;
	Server = new TIdTCPServer(0);
  Server->DefaultPort = port;
  Server->OnConnect = &ServerConnect;
  Server->OnDisconnect = &ServerDisconnect;
  Server->OnStatus = &ServerStatus;
  Server->OnExecute = &ServerExecute;
  Protocol = new TProtocol(this, connTOms, subsTOms, buffSubs);
  Tags = new TStringList();
}
//---------------------------------------------------------------------------
__fastcall TOPCGate::TOPCGate(int port, const AnsiString &certFile,
		const AnsiString &keyFile, const AnsiString &rootCertFile,
    TLogFlags logFlags, int connTOms, int subsTOms, bool buffSubs)
  :	LogFlags(logFlags)
{
	::LibInit();
	Intercept = new TServerInterceptOpenSSL(0);
  Intercept->OnStatusInfo = &InterceptStatus;
  Intercept->OnVerifyPeer = &InterceptVerifyPeer;
  Intercept->SSLOptions->CertFile = certFile;
  Intercept->SSLOptions->KeyFile = keyFile;
  Intercept->SSLOptions->RootCertFile = rootCertFile;
  Intercept->SSLOptions->Method = Sslopenssl::sslvSSLv23;
  Intercept->SSLOptions->Mode = Sslopenssl::sslmServer;
  Intercept->SSLOptions->VerifyMode = TSSLVerifyModeSet()
			<< Sslopenssl::sslvrfPeer
      << Sslopenssl::sslvrfFailIfNoPeerCert
      << Sslopenssl::sslvrfClientOnce;
	Server = new TIdTCPServer(0);
  Server->Intercept = Intercept;
  Server->DefaultPort = port;
  Server->OnConnect = &ServerConnect;
  Server->OnDisconnect = &ServerDisconnect;
  Server->OnStatus = &ServerStatus;
  Server->OnExecute = &ServerExecute;
  Protocol = new TProtocol(this, connTOms, subsTOms, buffSubs);
  Tags = new TStringList();
}
//---------------------------------------------------------------------------
__fastcall TOPCGate::~TOPCGate()
{
	delete Tags;
  delete Protocol;
  delete Server;
	delete Intercept;
}
//---------------------------------------------------------------------------
void __fastcall TOPCGate::Start()
{
	GetTags(Tags);
  TagsHash = SHA256(Tags->Text);
  Server->Active = true;
}
//---------------------------------------------------------------------------
void __fastcall TOPCGate::Stop()
{
	Server->Active = false;
}
//---------------------------------------------------------------------------
void __fastcall TOPCGate::InterceptStatus(AnsiString msg)
{
	if (LogFlags & lfIntercept)
		Log("InterceptStatus: " + msg);
}
//---------------------------------------------------------------------------
bool __fastcall TOPCGate::InterceptVerifyPeer(TX509 *cert, int ok)
{
	if (LogFlags & lfIntercept)
		Log("InterceptVerifyPeer: " + cert->Subject->OneLine + ", ok=" + ok);
	return true;
}
//---------------------------------------------------------------------------
void __fastcall TOPCGate::ServerExecute(TIdPeerThread *thread)
{
	if (LogFlags & lfServer)
		Log("ServerExecute");
  try {
	  while (Protocol->Execute(thread))
  		;
  } catch(EIdConnClosedGracefully &e) {
  } catch(Exception &e) {
		if (LogFlags & lfError)
			Log("ServerExecute Error: " + e.Message);
  }
	if (LogFlags & lfServer)
		Log("~ServerExecute");
  thread->Connection->Disconnect();
}
//---------------------------------------------------------------------------
void __fastcall TOPCGate::ServerConnect(TIdPeerThread *thread)
{
	if (LogFlags & lfServer)
		Log("ServerConnect");
  Server->Threads->LockList();
  try {
		thread->Data = new TSession();
  } __finally {
	  Server->Threads->UnlockList();
  }
}
//---------------------------------------------------------------------------
void __fastcall TOPCGate::ServerDisconnect(TIdPeerThread *thread)
{
	if (LogFlags & lfServer)
		Log("ServerDisconnect");
  Server->Threads->LockList();
  try {
    delete thread->Data;
    thread->Data = 0;
  } __finally {
	  Server->Threads->UnlockList();
  }
}
//---------------------------------------------------------------------------
void __fastcall TOPCGate::ServerStatus(TObject *sender, const TIdStatus sts,
 		const AnsiString msg)
{
	if (LogFlags & lfServer)
		Log("ServerStatus: " + msg);
}
//---------------------------------------------------------------------------
void __fastcall TOPCGate::SetValues(TStrings *sl)
{
  TList *tl = Server->Threads->LockList();
  try {
  	int nt = tl->Count;
    for (int i = 0; i < nt; ++i) {
      TIdPeerThread *thread = (TIdPeerThread*)tl->Items[i];
      if (thread->Data)
	     	Protocol->SendToSubscr(thread, sl);
    }
  } __finally {
	  Server->Threads->UnlockList();
  }
}
//---------------------------------------------------------------------------
/*
-   (  1)
- ()
-:
	b - bad
  g - good
-
-
  */                          
//static
AnsiString __fastcall TOPCGate::FormatVal(int idx, bool val, bool good,
 		const TDateTime &dt)
{
	return AnsiString(idx) + (val ? " 1" : " 0") + (good ? " g " : " b ") +
  		FormatDateTime("dd\".\"mm\".\"yyyy\" \"hh\":\"nn\":\"ss\".\"zzz", dt);
}
//---------------------------------------------------------------------------
//static
AnsiString __fastcall TOPCGate::FormatVal(int idx, char val, bool good,
		const TDateTime &dt)
{
	return AnsiString(idx) + " " + int(val) + (good ? " g " : " b ") +
  		FormatDateTime("dd\".\"mm\".\"yyyy\" \"hh\":\"nn\":\"ss\".\"zzz", dt);
}
//---------------------------------------------------------------------------
//static
AnsiString __fastcall TOPCGate::FormatVal(int idx, unsigned char val, bool good,
 		const TDateTime &dt)
{
	return AnsiString(idx) + " " + int(val) + (good ? " g " : " b ") +
  		FormatDateTime("dd\".\"mm\".\"yyyy\" \"hh\":\"nn\":\"ss\".\"zzz", dt);
}
//---------------------------------------------------------------------------
//static
AnsiString __fastcall TOPCGate::FormatVal(int idx, short val, bool good,
 		const TDateTime &dt)
{
	return AnsiString(idx) + " " + val + (good ? " g " : " b ") +
  		FormatDateTime("dd\".\"mm\".\"yyyy\" \"hh\":\"nn\":\"ss\".\"zzz", dt);
}
//---------------------------------------------------------------------------
//static
AnsiString __fastcall TOPCGate::FormatVal(int idx, unsigned short val, bool good,
 		const TDateTime &dt)
{
	return AnsiString(idx) + " " + val + (good ? " g " : " b ") +
  		FormatDateTime("dd\".\"mm\".\"yyyy\" \"hh\":\"nn\":\"ss\".\"zzz", dt);
}
//---------------------------------------------------------------------------
//static
AnsiString __fastcall TOPCGate::FormatVal(int idx, int val, bool good,
 		const TDateTime &dt)
{
	return AnsiString(idx) + " " + val + (good ? " g " : " b ") +
  		FormatDateTime("dd\".\"mm\".\"yyyy\" \"hh\":\"nn\":\"ss\".\"zzz", dt);
}
//---------------------------------------------------------------------------
//static
AnsiString __fastcall TOPCGate::FormatVal(int idx, unsigned int val, bool good,
 		const TDateTime &dt)
{
	return AnsiString(idx) + " " + val + (good ? " g " : " b ") +
  		FormatDateTime("dd\".\"mm\".\"yyyy\" \"hh\":\"nn\":\"ss\".\"zzz", dt);
}
//---------------------------------------------------------------------------
//static
AnsiString __fastcall TOPCGate::FormatVal(int idx, long val, bool good,
 		const TDateTime &dt)
{
	return AnsiString(idx) + " " + val + (good ? " g " : " b ") +
  		FormatDateTime("dd\".\"mm\".\"yyyy\" \"hh\":\"nn\":\"ss\".\"zzz", dt);
}
//---------------------------------------------------------------------------
//static
AnsiString __fastcall TOPCGate::FormatVal(int idx, unsigned long val, bool good,
 		const TDateTime &dt)
{
	return AnsiString(idx) + " " + val + (good ? " g " : " b ") +
  		FormatDateTime("dd\".\"mm\".\"yyyy\" \"hh\":\"nn\":\"ss\".\"zzz", dt);
}
//---------------------------------------------------------------------------
//static
AnsiString __fastcall TOPCGate::FormatVal(int idx, float val, bool good,
 		const TDateTime &dt)
{
	AnsiString v = val;
  if (DecimalSeparator != '.')
  	for (int i = 1, l = v.Length(); i <= l; ++i)
    	if (v[i] == DecimalSeparator) {
      	v[i] = '.';
        break;
      }
	return AnsiString(idx) + " " + v + (good ? " g " : " b ") +
  		FormatDateTime("dd\".\"mm\".\"yyyy\" \"hh\":\"nn\":\"ss\".\"zzz", dt);
}
//---------------------------------------------------------------------------
//static
AnsiString __fastcall TOPCGate::FormatVal(int idx, double val, bool good,
 		const TDateTime &dt)
{
	AnsiString v = val;
  if (DecimalSeparator != '.')
  	for (int i = 1, l = v.Length(); i <= l; ++i)
    	if (v[i] == DecimalSeparator) {
      	v[i] = '.';
        break;
      }
	return AnsiString(idx) + " " + v + (good ? " g " : " b ") +
  		FormatDateTime("dd\".\"mm\".\"yyyy\" \"hh\":\"nn\":\"ss\".\"zzz", dt);
}
//---------------------------------------------------------------------------
} //namespace OPCGate
//---------------------------------------------------------------------------

